#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>


int main( int argc, char** argv ) {

    int rank, size;
    int width;
    MPI_Comm cart_2D, row_comm;
    int dims[2], periods[2], remain_dims[2];
    int token;
    int local_size, local_rank;

    MPI_Init( &argc, &argv );
    MPI_Comm_size( MPI_COMM_WORLD, &size );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );

    width = atoi( argv[1] );

    if( size % width ) {
        fprintf( stderr, "The number of processes must be a multiple of the grid width (%d et %d)\n", size, width );
        MPI_Finalize();
        return EXIT_FAILURE;
    }

    /* Create a 2D cartesian topology */

    dims[0] = width;
    dims[1] = size/width;
    periods[0] = 0;
    periods[1] = 1;

    MPI_Cart_create( MPI_COMM_WORLD, 2, dims, periods, 0, &cart_2D );

    if( MPI_COMM_NULL == cart_2D ) {
        fprintf( stderr, "Error while creating the cartesian topology\n" );
        MPI_Comm_free( &cart_2D );
        MPI_Finalize();
        return EXIT_FAILURE;
    }

    /* Extract row communicators */

    remain_dims[0] = 1;
    remain_dims[1] = 0;

    MPI_Cart_sub( cart_2D, remain_dims, &row_comm );

    if( MPI_COMM_NULL == row_comm ) {
        fprintf( stderr, "Error while creating the row communicators\n" );
        MPI_Comm_free( &cart_2D );
        MPI_Comm_free( &row_comm );
        MPI_Finalize();
        return EXIT_FAILURE;
    }

    /* TODO: Extract column communicators */
    
    /* Do something on this communicator.... */
    /* TODO: Replace by your communications */

    MPI_Comm_size( row_comm, &local_size );
    MPI_Comm_rank( row_comm, &local_rank );

    if( 0 == local_rank ) {
        token = getpid();
    }

    MPI_Bcast( &token, 1, MPI_INT, 0, row_comm );

    fprintf( stdout, "%d/%d/%d : token = %d\n", rank, local_rank, size, token );

    /* Finalize: free the communicators we have created */

    MPI_Comm_free( &cart_2D );
    MPI_Comm_free( &row_comm );
    MPI_Finalize();
    return EXIT_SUCCESS;
}
