#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>

#define ROOT        0
#define DATASIZE    512
#define CHUNKSIZE   64

#define TAG_BEGIN   1
#define TAG_DATA    2
#define TAG_RESULT  4
#define TAG_END     8

int main( int argc, char** argv ) {

    int size, rank;
    int *data, *results, *local_data;
    int* tasks;
    int i, j, k;
    MPI_Status status;

    MPI_Init( &argc, &argv );
    MPI_Comm_size( MPI_COMM_WORLD, &size );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );

    /* Hello ! */

    fprintf( stderr, "I am rank %d in %d\n", rank, size );

    /* on prpare le tableau de donnes */

    if( ROOT == rank ) {
        data = (int *) malloc(  DATASIZE * DATASIZE * sizeof( int ) );
        for( i = 0 ; i <  DATASIZE ; ++i ) {
            for( j = 0 ; j < DATASIZE ; ++j ) {
                data[i * DATASIZE + j] = i * j;
            }
        }
    }

    if( ROOT == rank ) {
        results = (int *) malloc( DATASIZE * DATASIZE * sizeof( int ) );
    }

    local_data = (int *) malloc(  DATASIZE * (CHUNKSIZE + 2) * sizeof( int ) );

    /* ce tableau contiendra l'offset du chunk envoy  un worker */

    if( ROOT == rank ) {
        tasks = (int *) malloc( ( size - 1 ) * sizeof( int ) );
    }

    if( ROOT == rank ) {     /* MAITRE */

        int sender;
        int offset = 0;

        while( 1 ) {
            MPI_Recv( local_data, (CHUNKSIZE + 2) * DATASIZE, MPI_INT, 
                      MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
            
            sender = status.MPI_SOURCE;
            fprintf( stderr, "[ROOT] %d wants some work\n", sender );
            
            if( TAG_RESULT == status.MPI_TAG ) {
                /* rcupration du rsultat */
                memcpy( results + tasks[sender-1], local_data, 
                        CHUNKSIZE * DATASIZE * sizeof( int ) );
            }
            if( offset < DATASIZE * DATASIZE - CHUNKSIZE ) {            
                /* distribution du travail */
                if( 0 == offset ) {
                    for( i = 0 ; i < DATASIZE ; ++i ) {
                        local_data[i] = 0;
                    }
                    memcpy( &local_data[DATASIZE], &data[offset], DATASIZE * (CHUNKSIZE+1) * sizeof( int ) );
                } else {
                    if( offset >= DATASIZE * DATASIZE - (2 * CHUNKSIZE) ) { 
                        memcpy( local_data, &data[offset - CHUNKSIZE], DATASIZE * (CHUNKSIZE + 1) * sizeof( int ) );
                        for( i = 0 ; i < DATASIZE ; ++i ) {
                            local_data[i] = 0;
                        }
                    } else {
                        memcpy( local_data, &data[offset - CHUNKSIZE], DATASIZE * (CHUNKSIZE + 2) * sizeof( int ) );
                    }
                }

                MPI_Send( local_data, CHUNKSIZE * DATASIZE, MPI_INT, 
                          sender, TAG_DATA, MPI_COMM_WORLD );
                offset += CHUNKSIZE * DATASIZE;
                tasks[sender-1] = offset;
            } else {
                /* on rcupre tous les rsultats
                   et on signale aux esclaves qu'on a fini */
                MPI_Send( data, (CHUNKSIZE + 2) * DATASIZE, MPI_INT,
                          sender, TAG_END, MPI_COMM_WORLD );
                for( i = 2 ; i < size ; ++i ) {
                    MPI_Recv( local_data, CHUNKSIZE * DATASIZE, MPI_INT, 
                              MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
                    sender = status.MPI_SOURCE;
                    memcpy( results + tasks[sender-1], local_data, CHUNKSIZE * DATASIZE * sizeof( int ) );
                    MPI_Send( local_data, (CHUNKSIZE + 2) * DATASIZE, MPI_INT,
                              sender, TAG_END, MPI_COMM_WORLD );
                }
                fprintf( stderr, "[ROOT] done\n" );
                goto finish;        
            }
        }

    } else {                 /* ESCLAVES */
        /* on est prt  recevoir du travail donc on en demande */
        MPI_Send( local_data, CHUNKSIZE * DATASIZE, MPI_INT, 
                  ROOT, TAG_BEGIN, MPI_COMM_WORLD );

        while( 1 ) {
            /* rception de donnes  calculer */
            MPI_Recv( local_data, (CHUNKSIZE + 2) * DATASIZE, MPI_INT,
                      ROOT, MPI_ANY_TAG, MPI_COMM_WORLD, &status );

            /* est-ce vraiment du travail ? */
            if( TAG_DATA == status.MPI_TAG ) {
                fprintf( stderr, "[%d/%d] received some work\n", rank, size );
                /* au boulot ! */
                for( i = 1 ; i < CHUNKSIZE+2 ; ++i ) {
                    for( j = 1 ; j < CHUNKSIZE+2 ; ++j ) {
                        for( k = 1 ; k < CHUNKSIZE+2 ; ++k ) {
                            local_data[j*CHUNKSIZE + i] = local_data[(j-1)*CHUNKSIZE + i] + local_data[(j+1)*CHUNKSIZE + i];
                        }
                    }
                }
                /* on envoie le rsultat au maitre */
                MPI_Send( local_data, CHUNKSIZE * DATASIZE, MPI_INT,
                          ROOT, TAG_RESULT, MPI_COMM_WORLD );
            } else {
                /* on a fini */
                fprintf( stderr, "[%d] done\n", rank );
                goto finish;
            }
        }
    }
    
    finish:

    /* finalisation */
    
    if( ROOT == rank ) {
        //        free( data );
        free( results );
        free( tasks );
    }
    free( local_data );

    MPI_Finalize();
    return EXIT_SUCCESS;

}
