#include <stdio.h>
#include <stdlib.h>

#include <x86intrin.h>

#define VECSIZE 16

double rand_range_d( int min, int max ){
    return (double) min +  ( max - min + 1 )*((double)rand()/RAND_MAX);
}

int rand_range_i( int min, int max ){
    return min + (int) ( max - min + 1 )*((double)rand()/RAND_MAX);
}


/* Make sure this function is not vectorized */

__attribute__((optimize("no-tree-vectorize")))
void initmatrix( double * restrict matrix ){
    int i;
    for( i = 0 ; i < VECSIZE*VECSIZE ; i++ ) {
        matrix[i] = (double)i;
    }
}


int main( int argc, char** argv ){
    double matrix[VECSIZE*VECSIZE];
    double vector[VECSIZE];
    int i;

    __m256d a, b, res;
    __m256d c;
    double t0, t1, t2, t3;

    initmatrix( matrix );

    /* This is the interesting loop */
    
    for( i = 0 ; i < VECSIZE ; i+=4 ) {
        t0 = rand_range_d( 1, 4);
        t1 = rand_range_d( 1, 4);
        t2 = rand_range_d( 1, 4);
        t3 = rand_range_d( 1, 4);

        /* Load the data */
        
        a = _mm256_set_pd( matrix[(i+3+rand_range_i( 0, VECSIZE ))*VECSIZE],
                           matrix[(i+2+rand_range_i( 0, VECSIZE ))*VECSIZE],
                           matrix[(i+1+rand_range_i( 0, VECSIZE ))*VECSIZE],
                           matrix[(i+rand_range_i( 0, VECSIZE ))*VECSIZE] );
        b = _mm256_loadu_pd( &matrix[i*VECSIZE] );
        
        c = _mm256_set_pd( t3, t2, t1, t0 );

        /* Perform the operation */
        
        res = _mm256_add_pd( a, b );
        res = _mm256_mul_pd( res, c );

        /* Store the result */

        _mm256_storeu_pd( &vector[ i ], res );             
            
    }

    /* Necessary because with -O3, gcc detects that nothing is done */

    /*    for( i = 0 ; i < VECSIZE ; i++ ) {
        printf( "%.2lf ", vector[i] );
    }
    printf( "\n" );     */

   return EXIT_SUCCESS;
}
